# encoding: utf-8
from peewee import OperationalError, ProgrammingError
from playhouse.migrate import migrate
from pypi_server.db.migrator import migration


@migration(8)
def add_uniquie_basename_index(migrator, db):
    try:
        migrate(
            migrator.add_index('packagefile', ('basename',), True)
        )
    except (OperationalError, ProgrammingError):
        pass



#!/usr/bin/env python
# encoding: utf-8
from tornado.gen import coroutine
from tornado.web import HTTPError
from pypi_server.db.packages import PackageVersion, Package, PackageFile
from pypi_server.handlers import route
from pypi_server.handlers.api import JSONHandler
from pypi_server.handlers.api.login import authorization_required
from pypi_server.handlers.base import threaded
from pypi_server.handlers.pypi.proxy.client import normalize_package_name


@route('/api/v1/package/(?P<package>[\w\.\d\-\_]+)/(?P<version>[\w\d\.]+)/?')
class PackageVersionHandler(JSONHandler):
    @authorization_required()
    @coroutine
    def get(self, package, version):
        ver = yield self.get_version(package, version)
        self.response((yield self.version_info(ver)))

    @coroutine
    def put(self, package, version):
        ver = yield self.get_version(package, version)
        ver.hidden = self.json.get('hidden', ver.hidden)

        if ver.is_dirty:
            yield self.thread_pool.submit(ver.save)

        self.response((yield self.version_info(ver)))

    @authorization_required(is_admin=True)
    @coroutine
    def delete(self, package, version):
        ver = yield self.get_version(package, version)
        yield self.thread_pool.submit(ver.delete_instance, recursive=True)

    @staticmethod
    @threaded
    def get_version(package, version):
        package = normalize_package_name(package)
        q = PackageVersion.select().join(Package).join(PackageFile).where(
            Package.lower_name == package,
            PackageVersion.version == version,
        ).limit(1)

        if not q.count():
            raise HTTPError(404)

        return q[0]

    @staticmethod
    @threaded
    def version_info(version):
        return {
            "version": str(version.version),
            "hidden": version.hidden,
            "downloads": version.downloads,
            "author": version.author,
            "author_email": version.author_email,
            "maintainer": version.maintainer,
            "maintainer_email": version.maintainer_email,
            "home_page": version.home_page,
            "license": version.license,
            "summary": version.summary,
            "description": version.description,
            "keywords": version.keywords,
            "platform": version.platform,
            "download_url": version.download_url,
            "classifiers": version.classifiers,
            "requires": version.requires,
            "requires_dist": version.requires_dist,
            "provides": version.provides,
            "provides_dist": version.provides_dist,
            "requires_external": version.requires_external,
            "requires_python": version.requires_python,
            "obsoletes": version.obsoletes,
            "obsoletes_dist": version.obsoletes_dist,
            "project_url": version.project_url,
            "files": list(map(
                lambda f: dict(
                    name=f.basename,
                    url="/package/{0}/{1}/{2}".format(version.package.lower_name, version.version, f.basename)
                ),
                version.packagefile_set
            ))
        }



# encoding: utf-8
import os
import logging

from pypi_server.timeit import timeit
from tornado.gen import coroutine, Return
from tornado.ioloop import IOLoop
from tornado.web import HTTPError
from pypi_server.handlers import route, add_slash
from pypi_server.handlers.base import BaseHandler, threaded
from pypi_server.handlers.pypi.proxy.client import PYPIClient
from pypi_server.db.packages import PackageVersion, Package, PackageFile


log = logging.getLogger(__name__)


PACKAGE_META = {
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    "home_page": None,
    "license": None,
    "summary": None,
    "description": None,
    "keywords": None,
    "platform": None,
    "download_url": None,
    "classifiers": None,
    "requires": None,
    "requires_dist": None,
    "provides": None,
    "provides_dist": None,
    "requires_external": None,
    "requires_python": None,
    "obsoletes": None,
    "obsoletes_dist": None,
    "project_url": None,
}


def chunks(lst, n):
    for i in xrange(0, len(lst), n):
        yield lst[i:i + n]


@coroutine
def proxy_remote_package(package):
    pkg = yield threaded(Package.get_or_create)(name=package, proxy=True)

    releases, cached_releases = yield [PYPIClient.releases(pkg.name), threaded(pkg.versions)(True)]
    IOLoop.current().add_callback(threaded(pkg.hide_versions), filter(lambda x: not x.hidden, releases))

    cached_releases = set(map(lambda x: x.version, cached_releases))
    new_releases = list(releases - cached_releases)

    for release_part in chunks(new_releases, 10):
        yield [release_fetch(pkg, release) for release in release_part]

    raise Return(pkg)


@threaded
def release_db_save(package, rel, version_info, release_files):
    version = package.create_version(rel)
    version.fetched = False

    for key, val in PACKAGE_META.items():
        setattr(version, key, version_info.get(val if val else key))

    version.save()

    for f in release_files:
        pkg_file = version.create_file(f['filename'])
        pkg_file.fetched = False
        pkg_file.url = f['url']
        pkg_file.md5 = f['md5_digest']
        pkg_file.save()

    return version


@coroutine
def release_fetch(package, rel):
    version_info, release_files = yield [
        PYPIClient.release_data(package.name, rel),
        PYPIClient.release_files(package.name, rel)
    ]

    raise Return((yield release_db_save(package, rel, version_info, release_files)))


@route(r'/simple/(?P<package>[\w\.\d\-\_]+)/?')
@route(r'/simple/(?P<package>[\w\.\d\-\_]+)/(?P<version>[\w\.\d\-\_]+)/?')
@add_slash
class VersionsHandler(BaseHandler):
    @timeit
    @coroutine
    def get(self, package, version=None):
        exists, is_proxy, pkg = yield self.packages_list(package)

        if not exists or (is_proxy or not pkg):
           pkg = yield self.proxy_package(package)

        if not pkg:
            raise HTTPError(404)

        files = yield threaded(pkg.files)(version=version)

        if not files:
            raise HTTPError(404)

        self.render(
            os.path.join('simple', 'files.html'),
            package=pkg.lower_name,
            files=files
        )

    @threaded
    @timeit
    def packages_list(self, package):
        q = Package.select().join(
            PackageVersion
        ).join(
            PackageFile
        ).where(
            Package.lower_name == package.lower()
        ).limit(1)

        exists = q.count()
        is_proxy = q.where(Package.is_proxy == True).count()

        return exists, is_proxy, q[0] if exists else None

    @timeit
    @coroutine
    def proxy_package(self, package):
        if (yield PYPIClient.exists(package)):
            pkg_real_name = yield PYPIClient.find_real_name(package)
            pkg = yield proxy_remote_package(pkg_real_name)

            raise Return(pkg)

        raise LookupError("Remote package not found")



#!/usr/bin/env python
# encoding: utf-8
import os
from tornado.gen import coroutine
from pypi_server.handlers import route
from pypi_server.handlers.base import BaseHandler, threaded
from pypi_server.http_cache import HTTPCache
from pypi_server.cache import Cache
from pypi_server import ROOT


@route(r"^/$")
class IndexHander(BaseHandler):
    @HTTPCache(600, True, 60)
    @coroutine
    @Cache(60, ignore_self=True)
    def get(self):
        vendor_js, vendor_css, application, styles = yield [
            self.find(ROOT, 'static/vendor', '.js'),
            self.find(ROOT, 'static/vendor', '.css'),
            self.find(ROOT, 'static/js', '.js'),
            self.find(ROOT, 'static', 'style.css'),
        ]

        self.render(
            'index.html',
            vendor_js=vendor_js,
            vendor_css=vendor_css,
            application=application,
            styles=styles,
        )

    @staticmethod
    @threaded
    @Cache(60)
    def find(base_dir, path, pattern):
        file_list = []

        for root, dirs, files in os.walk(os.path.join(base_dir, *path.split('/'))):
            for f in filter(lambda x: x.endswith(pattern), files):
                rel_file = "/".join(os.path.split(os.path.join(root, f))).replace("/".join(os.path.split(base_dir)), '/')
                file_list.append("/{}".format(rel_file.strip("/")))

        return file_list


# encoding: utf-8
import logging
from functools import wraps
from time import time
from concurrent.futures import Future


log = logging.getLogger(__name__)


def timeit(func):
    def log_result(start_time):
        log.debug(
            'Time of execution function "%s": %0.6f',
            ".".join(filter(
                None,
                (
                    func.__module__,
                    func.__class__.__name__ if hasattr(func, '__class__') else None,
                    func.__name__
                )
            )),
            time() - start_time
        )

    @wraps(func)
    def wrap(*args, **kwargs):
        start_time = time()

        result = func(*args, **kwargs)
        if isinstance(result, Future):
            result.add_done_callback(lambda x: log_result(start_time))
        else:
            log_result(start_time)

        return result

    return wrap



# encoding: utf-8
from pypi_server.db.migrator import migration
from pypi_server.db.users import Users


@migration(2)
def create_default_user(migrator, db):
    Users(login='admin', password='admin', email="admin@example.net", is_admin=True).save()



# encoding: utf-8
from playhouse.migrate import migrate
import peewee
from pypi_server.db.migrator import migration
from pypi_server.db.packages import PackageVersion


@migration(9)
def change_license_field_type(migrator, db):
    if isinstance(db, peewee.SqliteDatabase):
        # SQLite has not length
        return

    try:
        migrate(migrator.drop_column('packageversion', 'license_old'))
    except:
        pass

    with db.transaction():
        migrate(
            migrator.rename_column('packageversion', 'license', 'license_old'),
            migrator.add_column("packageversion", 'license', PackageVersion.license),
        )

        db.execute_sql("UPDATE packageversion SET license = license_old")
        migrate(migrator.drop_column('packageversion', 'license_old'))



# encoding: utf-8
from peewee import DoesNotExist
from tornado.web import HTTPError
from pypi_server.db.users import Users
from pypi_server.handlers import route
from pypi_server.handlers.base import threaded
from pypi_server.handlers.api import JSONHandler
from pypi_server.handlers.api.login import authorization_required
from pypi_server.handlers.api.users import LOGIN_EXP, EMAIL_EXP


@route('/api/v1/user/(\d+)/?')
class UserHandler(JSONHandler):
    @authorization_required(is_admin=True)
    @threaded
    def get(self, uid):
        try:
            user = Users.get(id=uid)
        except DoesNotExist:
            raise HTTPError(404)

        self.response(dict(
            id=user.id,
            login=user.login,
            email=user.email,
            is_admin=user.is_admin,
            disabled=user.disabled,
            packages=list(
                map(
                    lambda x: x.name,
                    user.package_set
                )
            )
        ))

    @authorization_required(is_admin=True)
    @threaded
    def put(self, uid):
        try:
            user = Users.get(id=uid)
            user.login = self.json.get("login", user.login)
            user.email = self.json.get("email", user.email)
            user.is_admin = bool(self.json.get("is_admin", user.is_admin))
            user.password = self.json.get("password", user.password)

            assert user.password and len(user.password) > 3
            assert LOGIN_EXP.match(user.login)
            assert EMAIL_EXP.match(user.email)
        except (KeyError, AssertionError):
            raise HTTPError(400)
        except DoesNotExist:
            raise HTTPError(404)
        else:
            user.save()

            self.response({
                'id': user.id,
                'login': user.login,
                'email': user.email,
                'is_admin': user.is_admin,
            })

    @authorization_required(is_admin=True)
    @threaded
    def delete(self, uid):
        try:
            user = Users.get(id=uid)
            user.disabled = True
        except (KeyError, AssertionError):
            raise HTTPError(400)
        except DoesNotExist:
            raise HTTPError(404)
        else:
            user.save()



# encoding: utf-8
import logging
from copy import copy
from tornado.gen import coroutine, Return
from tornado.httpclient import AsyncHTTPClient
from tornado.ioloop import IOLoop
from tornado.locks import Lock
from tornado_xmlrpc.client import ServerProxy
from pypi_server.cache import Cache, HOUR
from pypi_server.hash_version import HashVersion


log = logging.getLogger(__name__)


def normalize_package_name(name):
    return name.lower().replace("_", "-")


class PYPIClient(object):
    CLIENT = None
    BACKEND = None
    THREAD_POOL = None
    INDEX = None
    XMLRPC = None
    LOCK = None

    @classmethod
    def configure(cls, backend, thread_pool):
        cls.CLIENT = AsyncHTTPClient(io_loop=IOLoop.current())
        cls.BACKEND = backend
        cls.THREAD_POOL = thread_pool
        cls.XMLRPC = ServerProxy(
            str(copy(backend)(path="/pypi")),
        )
        cls.LOCK = Lock()

    @classmethod
    @coroutine
    @Cache(24 * HOUR)
    def packages(cls):
        with (yield cls.LOCK.acquire()):
            index = dict(
                map(
                    lambda x: (x.lower().replace("_", "-"), x),
                    (yield cls.XMLRPC.list_packages())
                )
            )

            log.info("Remote PYPI index updated: %d packages", len(index))
            raise Return(index)

    @classmethod
    @coroutine
    @Cache(HOUR)
    def search(cls, names, descriptions, operator="or"):
        assert operator in ('or', 'and')
        result = yield cls.XMLRPC.search({'name': names, 'description': descriptions}, operator)
        raise Return(result)

    @classmethod
    @coroutine
    def exists(cls, name):
        try:
            real_name = yield cls.find_real_name(name)
        except LookupError:
            raise Return(False)

        releases = yield cls.releases(real_name)
        if not releases:
            raise Return(False)

        raise Return(True)

    @classmethod
    @coroutine
    def find_real_name(cls, name):
        name = normalize_package_name(name)

        packages = yield cls.packages()
        real_name = packages.get(name.lower())

        if real_name is None:
            raise LookupError("Package not found")

        raise Return(real_name)

    @classmethod
    @coroutine
    @Cache(HOUR)
    def releases(cls, name):
        process_versions = lambda x: set(map(HashVersion, x))

        all_releases, current_releases = yield [
            cls.XMLRPC.package_releases(name, True),
            cls.XMLRPC.package_releases(name)
        ]

        all_releases = process_versions(all_releases)
        current_releases = process_versions(current_releases)

        hidden_releases = all_releases - current_releases

        res = []
        for x in current_releases:
            x.hidden = False
            res.append(x)

        for x in hidden_releases:
            x.hidden = True
            res.append(x)

        raise Return(set(res))

    @classmethod
    @coroutine
    @Cache(4 * HOUR)
    def release_data(cls, name, version):
        info = yield cls.XMLRPC.release_data(str(name), str(version))
        raise Return(info)

    @classmethod
    @coroutine
    @Cache(4 * HOUR)
    def release_files(cls, name, version):
        info = yield cls.XMLRPC.release_urls(str(name), str(version))
        raise Return(sorted(info, key=lambda x: x['filename']))



# encoding: utf-8
from distutils.version import LooseVersion
from functools import total_ordering


@total_ordering
class HashVersion(LooseVersion):
    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        if not isinstance(other, HashVersion):
            other = HashVersion(str(other))

        return tuple(self.version) == tuple(other.version)

    def __gt__(self, other):
        if not isinstance(other, HashVersion):
            other = HashVersion(str(other))

        return tuple(self.version) > tuple(other.version)



# encoding: utf-8
import peewee
from pypi_server.db.packages import Package, PackageVersion, PackageFile
from pypi_server.db.users import Users
from pypi_server.handlers import route
from pypi_server.handlers.api import JSONHandler
from pypi_server.handlers.base import threaded
from pypi_server.handlers.api.login import authorization_required


@route('/api/v1/packages/?')
class PackagesHandler(JSONHandler):
    @authorization_required()
    @threaded
    def get(self):
        q = Package.select(
            Package,
            Users.login,
        ).join(
            Users, peewee.JOIN.LEFT_OUTER
        )

        self.response(
            list(
                map(
                    lambda x: dict(
                        name=x.name,
                        owner={
                            'id': x.owner_id,
                            'name': x.owner.login
                        },
                        is_proxy=x.is_proxy,
                    ),
                    q
                )
            )
        )


# encoding: utf-8
import os
from tornado.gen import coroutine, Return
from tornado_xmlrpc.handler import XMLRPCHandler
from pypi_server.handlers import route, add_slash
from pypi_server.handlers.base import BaseHandler, threaded
from pypi_server.handlers.pypi.proxy.client import PYPIClient
from pypi_server.db.packages import Package, PackageVersion, PackageFile


@route(r'/simple/?')
@add_slash
class PackagesHandler(BaseHandler, XMLRPCHandler):
    @coroutine
    def get(self):
        self.render(
            os.path.join('simple', 'packages.html'),
            packages=(yield self.pkg_list())
        )

    @threaded
    def pkg_list(self):
        return list(Package.select().order_by(Package.lower_name))

    @threaded
    def rpc_package_releases(self, package_name, show_hidden=False):
        package = Package.select().where(Package.name == package_name)
        return list(map(lambda x: str(x.name), package.versions()))

    @threaded
    def rpc_release_urls(self, package_name, version):
        raise Return(
            list(
                map(
                    lambda x: "/".join((version, x.name)),
                    PackageFile.select().join(
                        Package
                    ).join(
                        PackageVersion
                    ).where(
                        Package.name == package_name,
                        PackageVersion.version == version
                    )
                )
            )
        )

    @threaded
    def rpc_list_packages(self):
        return list(map(lambda x: x.name, Package.select()))

    @coroutine
    def rpc_search(self, query, operator='or'):
        results = []

        assert operator in ("and", "or"), "Operator must be 'and' or 'or'"

        names = tuple(query.get('name', []))
        descriptions = tuple(query.get('summary', []))

        async_queries = yield [
            self._search(names, descriptions, operator),
            PYPIClient.search(names, descriptions, operator)
        ]

        for res in async_queries:
            for result in res:
                results.append(result)

        raise Return(results)

    @threaded
    def _search(self, names, descriptions, operator):
        op_and = lambda x, y: x & y
        op_or = lambda x, y: x | y

        operator_func = ({'or': op_or, 'and': op_and}).get(operator.lower(), op_and)
        result = []

        packages = PackageVersion.select().join(Package).where(
            operator_func(
                Package.name in tuple(names),
                PackageVersion.description.contains(descriptions)
            )
        )

        for pkg in packages:
            pkg_dict = dict(
                name=pkg.package.name,
                version=str(pkg.version.name),
                summary=pkg.version.description,
                _pypi_ordering=1,
            )

            result.append(pkg_dict)

        return result



# encoding: utf-8
from functools import wraps
from peewee import DoesNotExist
from tornado.web import HTTPError
from tornado.ioloop import IOLoop
from tornado.gen import Return, coroutine, maybe_future
from pypi_server.cache import Cache
from pypi_server.db.users import Users
from pypi_server.handlers import route
from pypi_server.handlers.api import JSONHandler, threaded


SESSION_DAYS = 3
SESSION_SECS = SESSION_DAYS * 86400


@threaded
@Cache(60)
def find_user(uid, is_admin=False):
    cond = (
        Users.disabled == False,
        Users.id == uid,
    )
    if is_admin:
        cond += (Users.is_admin == is_admin,)

    q = Users.select().where(*cond)

    if not q.count():
        raise DoesNotExist("User doesn't exists")

    return q.limit(1)[0]


def reject(self, reason):
    self.clear_all_cookies()
    self.set_status(403)
    raise Return(self.response({
        'error': reason
    }))


def authorization_required(is_admin=False):
    def decorator(func):
        @wraps(func)
        @coroutine
        def wrap(self, *args, **kwargs):
            io_loop = IOLoop.current()

            uid, remote_ip, start_time = self.get_secure_cookie('session', (None, None, None))
            if not all((uid, remote_ip, start_time)):
                raise Return(reject(self, 'Not authorized'))

            if not (start_time + SESSION_SECS) > io_loop.time():
                raise Return(reject(self, 'Session expired'))

            self.current_user = yield find_user(uid, is_admin=is_admin)

            if self.current_user.disabled:
                raise Return(reject(self, 'User disabled'))

            result = yield maybe_future(func(self, *args, **kwargs))
            raise Return(result)

        return wrap
    return decorator


@route("/api/v1/login")
class LoginHandler(JSONHandler):
    @authorization_required()
    def get(self):
        raise Return(self.response({
            'login': self.current_user.login,
            'is_admin': self.current_user.is_admin,
        }))

    @threaded
    def post(self):
        try:
            login, password = self.json['login'], self.json['password']
        except KeyError:
            raise HTTPError(400)

        try:
            user = Users.check(login, password)
        except DoesNotExist:
            raise HTTPError(403)

        self.set_secure_cookie(
            'session',
            (
                user.id,
                self.request.remote_ip,
                IOLoop.current().time()
            ),
            SESSION_DAYS
        )

        return self.response({
            'login': user.login,
            'is_admin': user.is_admin,
        })



# encoding: utf-8
from functools import wraps
from tornado.web import RequestHandler


try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import ujson as json
except ImportError:
    import json


class BaseHandler(RequestHandler):
    _NULL = object()
    THREAD_POOL = None
    STORAGE = None

    @property
    def thread_pool(self):
        return self.THREAD_POOL

    @classmethod
    def threaded(cls, func):
        @wraps(func)
        def wrap(*args, **kwargs):
            return cls.THREAD_POOL.submit(func, *args, **kwargs)

        return wrap

    def get_secure_cookie(self, name, value=None, max_age_days=31, min_version=None):
        val = super(BaseHandler, self).get_secure_cookie(
            name, None, max_age_days=max_age_days, min_version=min_version
        )

        if val is None:
            return value

        return pickle.loads(val) if val else val

    def set_secure_cookie(self, name, value, expires_days=30, version=None, **kwargs):
        return super(BaseHandler, self).set_secure_cookie(
            name, pickle.dumps(value, protocol=2), expires_days=expires_days,
            version=version, **kwargs
        )


threaded = BaseHandler.threaded


#!/usr/bin/env python
# encoding: utf-8
import hashlib
import datetime
from functools import total_ordering

import os
import logging
import peewee as p
from multiprocessing import RLock
from playhouse.kv import JSONField
from playhouse import signals
from pypi_server.cache import Cache, HOUR
from pypi_server.timeit import timeit
from pypi_server.hash_version import HashVersion
from pypi_server.db import Model
from pypi_server.db.users import Users
from tornado.ioloop import IOLoop


try:
    import cPickle as pickle
except ImportError:
    import pickle


log = logging.getLogger("db.packages")


class File(object):
    __slots__ = ('__close_callbacks', '__file')

    def __init__(self, name, mode='rb'):
        self.__close_callbacks = set()
        self.__file = open(name, mode)

    def add_close_callback(self, cb):
        self.__close_callbacks.add(cb)

    def close(self):
        self.__file.close()

        for cb in list(self.__close_callbacks):
            try:
                cb(self.__file)
            except:
                pass

    def __enter__(self):
        return self.__file

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class VersionField(p.CharField):
    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = kwargs.pop('max_length', 25)
        super(VersionField, self).__init__(*args, **kwargs)

    def python_value(self, value):
        return HashVersion(value) if value else None

    def db_value(self, value):
        return super(VersionField, self).db_value(str(value))


class FileField(p.CharField):
    STORAGE = None

    def python_value(self, value):
        abs_path = os.path.join(self.STORAGE, value)
        return os.path.join(abs_path)

    def db_value(self, value):
        return super(FileField, self).db_value(
            str(os.path.relpath(value, self.STORAGE))
        )


class Package(Model):
    name = p.CharField(index=True)
    lower_name = p.CharField(index=True, unique=True)
    owner = p.ForeignKeyField(Users, null=True, index=True)
    stable_version = VersionField(null=True)
    is_proxy = p.BooleanField(default=False, null=False)

    def __str__(self):
        return "%s" % self.name

    @timeit
    def hide_versions(self, releases):
        versions = dict(
            map(
                lambda x: (x.version, x),
                list(PackageVersion.select().join(Package).where(PackageVersion.package == self))
            )
        )

        for version in releases:
            ver = versions.get(version, None)
            if not ver:
                continue

            ver.hidden = getattr(version, 'hidden', False)

            if ver.is_dirty():
                log.debug("Set hidden flag for version %s of package %s as: %s", ver, self, ver.hidden)
                ver.save()

    @timeit
    def files(self, version=None):
        if version:
            version = self.find_version(version)

        q = PackageFile.select(
            Package,
            PackageVersion,
            PackageFile
        ).join(
            PackageVersion
        ).join(
            Package
        ).where(
            Package.id == self.id
        )

        if version:
            q = q.where(PackageFile.version == version)
        else:
            q = q.where(PackageVersion.hidden == False)

        return sorted(
            q.order_by(Package.name.asc()),
            key=lambda x: x.version.version,
            reverse=True,
        )

    @timeit
    def versions(self, show_hidden=False):
        q = PackageVersion.select(
            Package,
            PackageVersion
        ).join(
            Package
        ).where(
            PackageVersion.package == self
        )

        if not show_hidden:
            q = q.where(PackageVersion.hidden == False)

        return sorted(
            q,
            key=lambda x: x.version,
            reverse=True
        )

    @timeit
    def find_version(self, name):
        return PackageVersion.get(
            PackageVersion.package == self,
            PackageVersion.version == HashVersion(name),
        )

    @timeit
    def create_version(self, name):
        q = PackageVersion.select(
            PackageVersion.id
        ).where(
            PackageVersion.package == self,
            PackageVersion.version == name
        ).limit(1)

        if q.count():
            return PackageVersion.get(id=q[0].id)

        version = PackageVersion(
            package=self,
            version=name if isinstance(name, HashVersion) else HashVersion(name),
        )

        version.save()

        return version

    @classmethod
    @timeit
    def get_or_create(cls, name, proxy=False):
        if not Package.select().where(Package.name == name).count():
            pkg = Package(name=name, lower_name=name.lower(), is_proxy=proxy, owner=None)
            pkg.save()
            return pkg

        return Package.get(name=name)

    @classmethod
    @timeit
    def find(cls, name):
        q = Package.select().join(
            PackageVersion
        ).join(
            PackageFile
        ).where(
            Package.lower_name == name.lower()
        ).limit(1)

        if q.count():
            return q[0]

        raise LookupError('Package not found')


@total_ordering
class PackageVersion(Model):
    package = p.ForeignKeyField(Package, index=True, null=False)
    version = VersionField(index=True, null=False)
    hidden = p.BooleanField(default=False, null=False, index=True)
    downloads = p.IntegerField(default=0, index=True)
    author = p.CharField(max_length=255, null=True)
    author_email = p.CharField(max_length=255, null=True)
    maintainer = p.CharField(max_length=255, null=True)
    maintainer_email = p.CharField(max_length=255, null=True)
    home_page = p.CharField(max_length=256, null=True)
    license = p.TextField(null=True)
    summary = p.CharField(max_length=255, null=True)
    description = p.TextField(null=True)
    keywords = JSONField(default=[])
    platform = p.CharField(max_length=255, null=True)
    download_url = p.CharField(max_length=255, null=True)
    classifiers = JSONField(default=[])
    requires = JSONField(default=[])
    requires_dist = JSONField(default=[])
    provides = JSONField(default=[])
    provides_dist = JSONField(default=[])
    requires_external = JSONField(default=[])
    requires_python = p.CharField(max_length=255, null=True)
    obsoletes = JSONField(default=[])
    obsoletes_dist = JSONField(default=[])
    project_url = p.CharField(max_length=255, null=True)

    def __hash__(self):
        return hash((self.package_id, self.version))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __gt__(self, other):
        assert isinstance(other, PackageVersion)

        a = ()
        a += (self.package_id,)
        a += self.version.version
        b = ()
        b += (other.package_id,)
        b += other.version.version

        return a > b

    @timeit
    def files(self):
        return PackageFile.select().where(
            PackageFile.package == self.package,
            PackageFile.version == self
        )

    @timeit
    def find_file(self, name):
        return PackageFile.select().where(PackageFile.basename == name)

    @timeit
    def create_file(self, name):
        name = os.path.basename(name)
        pkg_file = PackageFile(
            package=self.package,
            version=self,
            basename=name,
            file=os.path.join(
                PackageFile.file.STORAGE,
                self.package.name,
                str(self.version),
                name
            )
        )

        pkg_file.save()
        return pkg_file

    @timeit
    def update_metadata(self, metadata):
        print (metadata)

    def __str__(self):
        return "%s" % self.version


class PackageFile(Model):
    LOCK = RLock()
    CHUNK_SIZE = 2 ** 16

    file = FileField(index=True, max_length=255)
    url = p.CharField(max_length=255, null=True, default=None)
    fetched = p.BooleanField(default=False, index=True)
    basename = p.CharField(max_length=255, index=True, unique=True)
    ts = p.DateTimeField(null=True)
    md5 = p.CharField(max_length=32, null=True)
    size = p.IntegerField(null=True)
    package = p.ForeignKeyField(Package, index=True)
    version = p.ForeignKeyField(PackageVersion, index=True)

    @classmethod
    def set_storage(cls, path):
        cls.file.STORAGE = path

    @timeit
    def exists(self):
        return os.path.exists(self.file) and not os.path.isdir(self.file)

    @timeit
    def open(self, mode='rb'):
        dirname = os.path.dirname(self.file)

        with self.LOCK:
            if not os.path.exists(dirname):
                os.makedirs(dirname)

        f = File(self.file, mode)

        def metadata_update(f):
            stat = os.stat(self.file)
            self.ts = datetime.datetime.fromtimestamp(stat.st_mtime)
            self.size = stat.st_size

            md5 = hashlib.md5()
            with open(self.file, 'rb') as fl:
                data = fl.read(self.CHUNK_SIZE)
                md5.update(data)
                while data:
                    data = fl.read(self.CHUNK_SIZE)
                    md5.update(data)

                    self.md5 = md5.hexdigest()

            self.save()

        if any(x in mode for x in 'wa'):
            f.add_close_callback(metadata_update)

        return f

    def __str__(self):
        return "%s" % self.basename


def remove_file(f):
    log.info('Removing file "%s"', f.file)
    os.remove(f.file)


@signals.pre_delete(PackageFile)
def on_delete_file(model_class, instance):
    io_loop = IOLoop.current()
    if model_class is PackageFile:
        files = [instance]

    elif model_class is PackageVersion:
        files = PackageFile.select().join(PackageVersion).where(PackageFile.version == instance)

    elif model_class is Package:
        files = PackageFile.select().join(Package).where(PackageFile.package == instance)

    for f in files:
        io_loop.add_callback(remove_file, f)



# encoding: utf-8
import os
from tornado.gen import coroutine, maybe_future, Return
from tornado.web import RedirectHandler, StaticFileHandler
from pypi_server import ROOT
from pypi_server.handlers import base


ROUTES = [
    (r"^/favicon.ico$", RedirectHandler, {'url': '/static/favicon/favicon.ico'}),
    (r"^/static/(.*)$", StaticFileHandler, {'path': os.path.join(ROOT, 'static')})
]


def route(uri, kwargs=None):
    kwargs = kwargs or {}

    def decorator(cls):
        ROUTES.append((uri, cls, kwargs))
        return cls

    return decorator


def add_slash(cls):
    def redirect(self, *args, **kwargs):
        pass

    class WrappedClass(cls):
        @coroutine
        def prepare(self, *args, **kwargs):
            if not self.request.path.endswith('/'):
                raise Return(self.redirect("{0}/".format(self.request.path)))
            else:
                raise Return((yield maybe_future(cls.prepare(self, *args, **kwargs))))

    WrappedClass.__name__ = cls.__name__

    return WrappedClass


from pypi_server.handlers.default import DefaultHandler
import pypi_server.handlers.index
import pypi_server.handlers.pypi
import pypi_server.handlers.api



# encoding: utf-8
import datetime
from functools import wraps


class HTTPCache(object):
    def __init__(self, timeout, use_expires=False, expire_timeout=60):
        self.timeout = timeout
        self.expire_timeout = expire_timeout
        self.use_expires = use_expires

    def __call__(self, func):
        @wraps(func)
        def wrap(handler, *args, **kwargs):
            ret = func(handler, *args, **kwargs)
            self.set_cache(handler)
            return ret

        return wrap

    def set_cache(self, handler, **kwargs):
        if hasattr(handler, '_new_cookie') and handler._new_cookie:
            return

        if handler._status_code == 200:
            if handler.request.method == "GET" and kwargs.get('timeout', self.timeout):
                handler.set_header("X-Accel-Expires", kwargs.get('timeout', self.timeout))

            if kwargs.get('use_expires', self.use_expires):
                handler.set_header(
                    "Expires",
                    (datetime.datetime.now() + datetime.timedelta(
                        seconds=kwargs.get('expire_timeout', self.expire_timeout)
                    )).strftime("%a, %d %b %Y %H:%M:%S %Z")
                )
                handler.set_header(
                    "Cache-Control",
                    "max-age={0}".format(kwargs.get('expire_timeout', self.expire_timeout))
                )


# encoding: utf-8
from pypi_server.db.migrator import MIGRATIONS, log


def migrate_db(DB, Migrations, migrator):
    migrations = sorted(MIGRATIONS, key=lambda x: x[0])

    q = Migrations.select(Migrations.id).order_by(Migrations.id.desc())

    if q.count():
        last_applied_migration = q[0].id
    else:
        last_applied_migration = -1

    for migration_id, name, migration in filter(lambda x: x[0] > last_applied_migration, migrations):
        tran = DB.transaction()

        try:
            log.info('Applying migration: "%s"', name)
            migration(migrator, DB)
            Migrations.create(
                id=migration_id,
                name=name
            )
        except Exception as e:
            log.error("Migration failed.")
            log.exception(e)
            tran.rollback()
            raise
        else:
            tran.commit()
